@@ -4,7 +4,7 @@ import functools
 import operator
 
 from .bindings import ffi, C, GSS_ERROR, _buf_to_str
-from .error import GSSCException, GSSException, GSSMechException
+from .error import GSSException, _exception_for_status
 from .names import MechName, Name
 from .oids import OID
 from .creds import Credential
@@ -173,9 +173,9 @@ class Context(object):
         try:
             if GSS_ERROR(retval):
                 if minor_status[0] and self.mech_type:
-                    raise GSSMechException(retval, minor_status[0], self.mech_type)
+                    raise _exception_for_status(retval, minor_status[0], self.mech_type)
                 else:
-                    raise GSSCException(retval, minor_status[0])
+                    raise _exception_for_status(retval, minor_status[0])
 
             output_token = _buf_to_str(output_token_buffer[0])
             return output_token
@@ -222,9 +222,9 @@ class Context(object):
         )
         if GSS_ERROR(retval):
             if minor_status[0] and self.mech_type:
-                raise GSSMechException(retval, minor_status[0], self.mech_type)
+                raise _exception_for_status(retval, minor_status[0], self.mech_type)
             else:
-                raise GSSCException(retval, minor_status[0])
+                raise _exception_for_status(retval, minor_status[0])
         return qop_state[0]
 
     def wrap(self, message, conf_req=True, qop_req=C.GSS_C_QOP_DEFAULT):
@@ -271,9 +271,9 @@ class Context(object):
         try:
             if GSS_ERROR(retval):
                 if minor_status[0] and self.mech_type:
-                    raise GSSMechException(retval, minor_status[0], self.mech_type)
+                    raise _exception_for_status(retval, minor_status[0], self.mech_type)
                 else:
-                    raise GSSCException(retval, minor_status[0])
+                    raise _exception_for_status(retval, minor_status[0])
 
             output_token = _buf_to_str(output_token_buffer[0])
             if conf_req and not conf_state[0]:
@@ -323,9 +323,9 @@ class Context(object):
         try:
             if GSS_ERROR(retval):
                 if minor_status[0] and self.mech_type:
-                    raise GSSMechException(retval, minor_status[0], self.mech_type)
+                    raise _exception_for_status(retval, minor_status[0], self.mech_type)
                 else:
-                    raise GSSCException(retval, minor_status[0])
+                    raise _exception_for_status(retval, minor_status[0])
 
             output = _buf_to_str(output_buffer[0])
             if conf_req and not conf_state[0]:
@@ -364,9 +364,9 @@ class Context(object):
         )
         if GSS_ERROR(retval):
             if minor_status[0] and self.mech_type:
-                raise GSSMechException(retval, minor_status[0], self.mech_type)
+                raise _exception_for_status(retval, minor_status[0], self.mech_type)
             else:
-                raise GSSCException(retval, minor_status[0])
+                raise _exception_for_status(retval, minor_status[0])
 
         return max_input_size[0]
 
@@ -397,9 +397,9 @@ class Context(object):
         try:
             if GSS_ERROR(retval):
                 if minor_status[0] and self.mech_type:
-                    raise GSSMechException(retval, minor_status[0], self.mech_type)
+                    raise _exception_for_status(retval, minor_status[0], self.mech_type)
                 else:
-                    raise GSSCException(retval, minor_status[0])
+                    raise _exception_for_status(retval, minor_status[0])
 
             exported_token = _buf_to_str(output_token_buffer[0])
             # Set our context to a 'blank' context
@@ -434,7 +434,7 @@ class Context(object):
         )
         try:
             if GSS_ERROR(retval):
-                raise GSSCException(retval, minor_status[0])
+                raise _exception_for_status(retval, minor_status[0])
 
             src_name = ffi.new('gss_name_t[1]')
             target_name = ffi.new('gss_name_t[1]')
@@ -456,7 +456,7 @@ class Context(object):
             src_name = Name(src_name)
             target_name = Name(target_name)
             if GSS_ERROR(retval):
-                raise GSSCException(retval, minor_status[0])
+                raise _exception_for_status(retval, minor_status[0])
 
             mech = OID(mech_type[0][0]) if mech_type[0] else None
 
@@ -503,9 +503,9 @@ class Context(object):
         )
         if GSS_ERROR(retval):
             if minor_status[0] and self.mech_type:
-                raise GSSMechException(retval, minor_status[0], self.mech_type)
+                raise _exception_for_status(retval, minor_status[0], self.mech_type)
             else:
-                raise GSSCException(retval, minor_status[0])
+                raise _exception_for_status(retval, minor_status[0])
         return lifetime_rec[0]
 
     def delete(self):
@@ -536,9 +536,9 @@ class Context(object):
         try:
             if GSS_ERROR(retval):
                 if minor_status[0] and self.mech_type:
-                    raise GSSMechException(retval, minor_status[0], self.mech_type)
+                    raise _exception_for_status(retval, minor_status[0], self.mech_type)
                 else:
-                    raise GSSCException(retval, minor_status[0])
+                    raise _exception_for_status(retval, minor_status[0])
 
             return _buf_to_str(output_token_buffer[0])
         finally:
@@ -549,7 +549,7 @@ class Context(object):
 class InitContext(Context):
     """
     An instance of this class can be used to initiate a secure context between two applications.
-    The application using `InitContext` must specify the :class:`~gssapi.names.Name` of the
+    The application using :class:`InitContext` must specify the :class:`~gssapi.names.Name` of the
     acceptor it wants to connect to, and send the first token it obtains from :meth:`step` to the
     acceptor. The acceptor should then respond and proceed with the establishment of the security
     context.
@@ -570,7 +570,9 @@ class InitContext(Context):
     :param time_req: the number of seconds for which the context should be valid. If not provided,
         a default lifetime will be used.
     :type time_req: int
-
+    :param input_chan_bindings: Optional channel bindings object, to bind this security context to
+        an underlying communications channel.
+    :type input_chan_bindings: :class:`~gssapi.chanbind.ChannelBindings`
     """
 
     def __init__(self, peer_name, cred=C.GSS_C_NO_CREDENTIAL, mech_type=None, req_flags=(), time_req=0,
@@ -592,18 +594,28 @@ class InitContext(Context):
         self._desired_mech = mech_type
         self._req_flags = functools.reduce(operator.or_, req_flags, 0)
         self._time_req = time_req
-        self._input_chan_bindings = ffi.cast('gss_channel_bindings_t', input_chan_bindings)
+
+        if (
+            hasattr(input_chan_bindings, '_cb')
+            and isinstance(input_chan_bindings._cb, ffi.CData)
+            and ffi.typeof(input_chan_bindings._cb) == ffi.typeof('gss_channel_bindings_t')
+        ):
+            self._channel_bindings = input_chan_bindings._cb
+        else:
+            self._channel_bindings = ffi.cast(
+                'gss_channel_bindings_t', C.GSS_C_NO_CHANNEL_BINDINGS
+            )
 
     def step(self, input_token=None):
         """Performs a step to establish the context as an initiator.
 
-        This method should be called in a loop and fed input tokens
-        from the acceptor, and its output tokens should be sent to the
-        acceptor, until this context's established attribute is True.
+        This method should be called in a loop and fed input tokens from the acceptor, and its
+        output tokens should be sent to the acceptor, until this context's :attr:`established`
+        attribute is True.
 
         :param input_token: The input token from the acceptor (omit this param or pass None on
             the first call).
-        :type input_token: bytes.
+        :type input_token: bytes
         :returns: either a byte string with the next token to send to the acceptor,
             or None if there is no further token to send to the acceptor.
         :raises: GSSException
@@ -614,8 +626,8 @@ class InitContext(Context):
         if input_token:
             input_token_buffer = ffi.new('gss_buffer_desc[1]')
             input_token_buffer[0].length = len(input_token)
-            c_str_import_token = ffi.new('char[]', input_token)
-            input_token_buffer[0].value = c_str_import_token
+            c_str_input_token = ffi.new('char[]', input_token)
+            input_token_buffer[0].value = c_str_input_token
         else:
             input_token_buffer = ffi.cast('gss_buffer_t', C.GSS_C_NO_BUFFER)
 
@@ -628,6 +640,7 @@ class InitContext(Context):
         output_token_buffer = ffi.new('gss_buffer_desc[1]')
         actual_flags = ffi.new('OM_uint32[1]')
         actual_time = ffi.new('OM_uint32[1]')
+
         if self._cred_object is not None:
             cred = self._cred_object._cred[0]
         else:
@@ -641,7 +654,7 @@ class InitContext(Context):
             desired_mech,
             self._req_flags,
             self._time_req,
-            self._input_chan_bindings,
+            self._channel_bindings,
             input_token_buffer,
             actual_mech,
             output_token_buffer,
@@ -656,9 +669,9 @@ class InitContext(Context):
 
             if GSS_ERROR(retval):
                 if minor_status[0] and actual_mech[0]:
-                    raise GSSMechException(retval, minor_status[0], actual_mech[0], out_token)
+                    raise _exception_for_status(retval, minor_status[0], actual_mech[0], out_token)
                 else:
-                    raise GSSCException(retval, minor_status[0], out_token)
+                    raise _exception_for_status(retval, minor_status[0], None, out_token)
 
             self.established = not (retval & C.GSS_S_CONTINUE_NEEDED)
             self.flags = actual_flags[0]
@@ -687,10 +700,13 @@ class AcceptContext(Context):
     context on the acceptor side. The initiator is normally authenticated as part of the context
     establishment process, though some mechanisms support anonymous peers.
 
-    :param cred: The credential to use for the acceptor. Pass :const:`gssapi.C_NO_CREDENTIAL` to
-        use the default acceptor credentials (e.g. any principal in the default keytab, when the
-        Kerberos mechanism is used).
+    :param cred: The credential to use for the acceptor. Omit this parameter or pass
+        :const:`gssapi.C_NO_CREDENTIAL` to use the default acceptor credentials (e.g. any principal
+        in the default keytab, when the Kerberos mechanism is used).
     :type cred: :class:`~gssapi.creds.Credential`
+    :param input_chan_bindings: Optional channel bindings object, to bind this security context to
+        an underlying communications channel.
+    :type input_chan_bindings: :class:`~gssapi.chanbind.ChannelBindings`
 
     .. py:attribute:: delegated_cred
 
@@ -721,14 +737,21 @@ class AcceptContext(Context):
         self.delegated_cred = None
         self.peer_name = None
 
-        self._input_chan_bindings = ffi.cast('gss_channel_bindings_t', input_chan_bindings)
+        if (
+            hasattr(input_chan_bindings, '_cb')
+            and isinstance(input_chan_bindings._cb, ffi.CData)
+            and ffi.typeof(input_chan_bindings._cb) == ffi.typeof('gss_channel_bindings_t')
+        ):
+            self._channel_bindings = input_chan_bindings._cb
+        else:
+            self._channel_bindings = ffi.cast('gss_channel_bindings_t', C.GSS_C_NO_CHANNEL_BINDINGS)
 
     def step(self, input_token):
         """Performs a step to establish the context as an acceptor.
 
-        This method should be called in a loop and fed input tokens
-        from the initiator, and its output tokens should be sent to the
-        initiator, until this context's established attribute is True.
+        This method should be called in a loop and fed input tokens from the initiator, and its
+        output tokens should be sent to the initiator, until this context's :attr:`established`
+        attribute is True.
 
         :param input_token: The input token from the initiator (required).
         :type input_token: bytes
@@ -759,7 +782,7 @@ class AcceptContext(Context):
             self._ctx,
             cred,
             input_token_buffer,
-            self._input_chan_bindings,
+            self._channel_bindings,
             src_name_handle,
             mech_type,
             output_token_buffer,
@@ -777,9 +800,9 @@ class AcceptContext(Context):
 
             if GSS_ERROR(retval):
                 if minor_status[0] and mech_type[0]:
-                    raise GSSMechException(retval, minor_status[0], mech_type[0], out_token)
+                    raise _exception_for_status(retval, minor_status[0], mech_type[0], out_token)
                 else:
-                    raise GSSCException(retval, minor_status[0], out_token)
+                    raise _exception_for_status(retval, minor_status[0], None, out_token)
 
             self.established = not (retval & C.GSS_S_CONTINUE_NEEDED)
             self.flags = actual_flags[0]
